# -*- coding: utf-8 -*-
# Copyright (C) 2017  IRISA
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# The original code contained here was initially developed by:
#
#     Pierre Vignet.
#     IRISA
#     Dyliss team
#     IRISA Campus de Beaulieu
#     35042 RENNES Cedex, FRANCE
"""
Search Minimal Accessibility Conditions

Simulation of the system until some halting condition (given with the final
property) is satisfied.
"""
from __future__ import unicode_literals
from __future__ import print_function

#from pycallgraph import PyCallGraph
#from pycallgraph.output import GraphvizOutput

# Standard imports
import os
from functools import partial
import sys
import itertools as it

# Multiprocessing
try:
    from concurrent.futures import ProcessPoolExecutor, as_completed
except ImportError:
    raise ImportError(
        "No module named concurrent.futures.\n"
        "You can try to install 'futures' module in Python 2.7"
    )
import multiprocessing as mp

# Custom imports
from cadbiom.models.clause_constraints.mcl.MCLAnalyser import MCLAnalyser
from cadbiom.models.clause_constraints.mcl.MCLQuery import MCLSimpleQuery
import cadbiom.commons as cm

LOGGER = cm.logger()


class ErrorRep(object):
    """Cf class CompilReporter(object):
    gt_gui/utils/reporter.py
    """

    def __init__(self):
        self.context = ""
        self.error = False

    def display(self, mess):
        self.error = True
        LOGGER.error(">> Context: %s; %s", self.context, mess)
        exit()

    def display_info(self, mess):
        LOGGER.error("-- Context: %s; %s", self.context, mess)
        exit()

    def set_context(self, cont):
        self.context = cont


def logical_operator(elements, operator):
    """Join elements with the given logical operator.

    :param arg1: Iterable of elements to join with a logical operator
    :param arg2: Logical operator to use 'and' or 'or'
    :return: logical_formula: str - AND/OR of the input list
    :type arg1: <list>
    :type arg2: <str>
    :rtype: <str>
    """
    assert operator in ("and", "or")
    #    print(operator + ": elements:", elements)

    return "(" + " {} ".format(operator).join(elements) + ")"


def make_logical_formula(previous_frontier_places, start_prop):
    """Make a logical formula based on previous results of MAC.

    The aim is to exclude previous solution.

    - 1 line: ``"A B" => (A and B)``
    - another line: ``"B C" => (B and C)``
    - merge all lines: ``(A and B) or (B and C)``
    - forbid all combinaisons: ``not((A and B) or (B and C))``

    :param arg1: Set of previous frontier places (previous solutions).
    :param arg2: Original property (constraint) for the solver.
    :return: A logical formula which excludes all the previous solutions.
    :type arg1: <set>
    :type arg2: <str>
    :rtype: <str>
    """

    logical_and = partial(logical_operator, operator="and")
    logical_or = partial(logical_operator, operator="or")

    def add_start_prop(prev_frontier_places_formula):
        """Deal with start_prop if given"""

        if start_prop and prev_frontier_places_formula:
            return start_prop + " and (" + prev_frontier_places_formula + ")"
        elif prev_frontier_places_formula:
            return prev_frontier_places_formula
        return start_prop

    mac_list = [
        logical_and(frontier_places) for frontier_places in previous_frontier_places
    ]

    if mac_list:
        # Logical or between each line
        return add_start_prop("not(" + logical_or(mac_list) + ")")
    return add_start_prop("")


def get_dimacs_start_properties(mcla, previous_frontier_places):
    """Translate frontier places to their numerical values
    thanks to the current unfolder.

    It's much more efficient than using the ANTLR grammar to parse formulas
    for each new query.

    :return: List of previous solutions (list of negative values of frontier places)
        Ex: [[-1, -2], [-2, -3], ...]
    :rtype: <list <list <int>>
    """
    dimacs_start = [
        [-mcla.unfolder.var_dimacs_code(place) for place in frontier_places]
         for frontier_places in previous_frontier_places
    ]
    return dimacs_start


def search_entry_point(
    model_file, mac_file, mac_step_file, mac_complete_file, mac_strong_file,
    steps, final_prop, start_prop, inv_prop, all_macs, continue_run, limit):
    """Search solutions

    :param model_file: Model file (bcx, xml, cal).
    :param mac_file: File used to store Minimal Activation Condition (MAC/CAM).
    :param mac_step_file: File used to store Minimal step numbers for each solution.
    :param mac_complete_file: File used to store MAC & trajectories.
    :param mac_strong_file: ???
    :param steps: Maximal steps to reach the solutions.
    :param final_prop: Formula: Property that the solver looks for.
    :param start_prop: Formula: Property that will be part of the initial state
        of the model.
        In concrete terms, some entities can be activated by this mechanism
        without modifying the model.
    :param inv_prop: Formula: Invariant property that will always occur during
        the simulation.
        The given logical formula will be checked at each step of the simulation.
    :param all_macs: If set to True (not default), search all macs with
        less or equal the maxium of steps defined with the argument `steps`.
        If set to False: The solver will search all solutions with the minimum
        of steps found in the first returned solution.

        :Example:

            all_macs = False, steps = 10;
            First solution found with 4 steps;
            The next solution will be searched with a maximum of 4 steps;

            all_macs = True, steps = 10;
            First solution found with 4 steps;
            The next solution is not reachable with 4 steps but with 5 steps
            (which is still less than 10 steps);
            Get the solution for 5 steps;

    :param continue_run: If set to True (not default), previous macs from a previous
        run, will be reloaded.
    :param limit: Limit the number of solutions.
    :type model_file: <str>
    :type mac_file: <str>
    :type mac_step_file: <str>
    :type mac_complete_file: <str>
    :type mac_strong_file: <str>
    :type steps: <int>
    :type final_prop: <str>
    :type start_prop: <str>
    :type inv_prop: <str>
    :type all_macs: <boolean>
    :type continue_run: <boolean>
    :type limit: <int>
    """
    # Build MCLA with Error Reporter
    mcla = MCLAnalyser(ErrorRep())

    # Load model file
    detect_model_type(mcla, model_file)(model_file)
    if mcla.reporter.error:
        raise "Error during loading of file"

    # Frontier places asked
    if continue_run:
        # Reload previous working files
        try:
            ## TODO: see docstring of read_mac_file
            previous_frontier_places = read_mac_file(mac_file)
            LOGGER.info(
                "%s:: Reload previous frontier places: %s",
                final_prop,
                len(previous_frontier_places),
            )
        except IOError:
            LOGGER.warning("%s:: mac file not found!", final_prop)
            previous_frontier_places = set()
    else:
        # New run
        previous_frontier_places = set()

    # Compute the number of previously computed frontier places
    current_nb_sols = len(previous_frontier_places) if previous_frontier_places else 0

    if current_nb_sols >= limit:
        # EXIT
        LOGGER.info(
            "%s:: Reaching the limitation of the number of solutions!", final_prop
        )
        return

    ## Find macs in an optimized way with find_macs() (see the doc for more info)
    LOGGER.info("%s:: Start property: %s", final_prop, start_prop)
    # with PyCallGraph(output=GraphvizOutput()):
    find_macs(mcla,
              mac_file, mac_step_file, mac_complete_file,
              steps, final_prop, start_prop, inv_prop, limit, current_nb_sols,
              previous_frontier_places)

    return
    ############################################################################
    ############################################################################
    ## Alternative and deprecated less efficient way
    ## Find mac one by one with find_mac() (see the docstring for more info)
    while True:
        # Compute the formula of the next start_property
        current_start_prop = make_logical_formula(previous_frontier_places, start_prop)
        LOGGER.info("%s:: Start property: %s", final_prop, current_start_prop)

        ret = \
            find_mac(mcla,
                     mac_file, mac_step_file, mac_complete_file,
                     steps, final_prop, start_prop, inv_prop,
                     previous_frontier_places)

        if not ret:
            # No new solution/not satisfiable
            return

        frontier_places, min_steps = ret

        # Add theese frontier places to set of previous ones
        # (tuple is hashable)
        previous_frontier_places.add(tuple(frontier_places))
        LOGGER.debug(
            "%s:: Prev frontier places: %s", final_prop, previous_frontier_places
        )

        # If set to True (not default), search all macs with
        # less or equal the maxium of steps defined with the argument `steps`.
        # If all_macs == True: The limit is always the maximal number given
        # in 'steps' argument.
        if not all_macs:
            steps = min_steps

        # Stop when limitation is reached
        current_nb_sols += 1
        if current_nb_sols >= limit:
            # EXIT
            LOGGER.info(
                "%s:: Reaching the limitation of the number of solutions!", final_prop
            )
            return

        LOGGER.debug("%s:: Next solution will be in %s steps", final_prop, steps)


def find_macs(mcla,
              mac_file, mac_step_file, mac_complete_file,
              steps, final_prop, start_prop, inv_prop, limit, current_nb_sols,
              previous_frontier_places):
    """Search for many solutions, save timings, and save frontiers.

    For every new solution, the system is NOT reinitialized, and a satisfiability
    test is made ONLY when there is no solution for the current step.
    This test is made to evaluate the minimal number of steps for reachability.

    Unlike find_mac(), this function is autonomous and takes into account the
    limitation of the number of solutions.

    .. TODO:: Handle all_macs flag like the old method with find_mac()
        Not used very often but can be usefull sometimes...

    :param limit: Limit the number of solutions.
    :param current_nb_sols: The current number of solutions already found.
        This number is used to limit the number of searched solutions.
    :param previous_frontier_places: Set of frontier places tuples from previous
        solutions. These tuples will be banned from the future solutions.
    :type limit: <int>
    :type current_nb_sols: <int>
    :type previous_frontier_places: <set <tuple <str>>>
    :return: None
    """
    # Build query
    query = MCLSimpleQuery(start_prop, inv_prop, final_prop)
    # Translate frontier places to their numerical values
    query.dim_start = get_dimacs_start_properties(mcla, previous_frontier_places)

    # Iterate on solutions
    for i, frontier_sol in enumerate(mcla.mac_search(query, steps), current_nb_sols):
        LOGGER.info("%s:: Current solution n°%s", final_prop, i + 1)

        LOGGER.debug("%s:: Next MAC object:\n%s", final_prop, frontier_sol)

        # Save MAC and timings
        LOGGER.debug("%s:: Save MAC and timings...", final_prop)
        with open(mac_complete_file, "a") as file:
            frontier_sol.save(file)

        # Save MAC
        LOGGER.debug("%s:: Save next MAC: %s", final_prop)
        with open(mac_file, "a") as file:
            frontier_sol.save(file, only_macs=True)

        # Save min steps
        # min_step = mcla.unfolder.current_step - 1 # Magic number !
        # LOGGER.debug("%s:: Save minimal steps: %s", final_prop, min_step)
        # with open(mac_step_file, 'a') as file:
        #     file.write(str(min_step)+'\n')

        # Stop when limitation is reached
        if i + 1 >= limit:
            # EXIT
            LOGGER.info(
                "%s:: Reaching the limitation of the number of solutions!", final_prop
            )
            return

    LOGGER.info("%s:: STOP the search! No more MAC.", final_prop)


def find_mac(mcla,
             mac_file, mac_step_file, mac_complete_file,
             steps, final_prop, start_prop, inv_prop, previous_frontier_places):
    """Search for 1 solution, save timings, save frontiers, and return it with
    the current step (deprecated, see find_macs()).

    For every new solution, the system is reinitialized, and a satisfiability
    test is made on a new query to evaluate the minimal number of steps for
    reachability.

    The side effect is that this process is expensive in a general way,
    and that parsing the properties (logical formulas of the frontier places of
    the previous solutions for example) in text format is very expensive because
    realized by the grammar ANTLR.

    :param previous_frontier_places: Set of frontier places tuples from previous
        solutions. These tuples will be banned from the future solutions.
    :type previous_frontier_places: <set <tuple <str>>>
    :return: A tuple of activated frontiers and the current step.
        None if there is no new Solution or if problem is not satisfiable.
    """
    # Build query
    query = MCLSimpleQuery(start_prop, inv_prop, final_prop)
    # Translate frontier places to their numerical values
    query.dim_start = get_dimacs_start_properties(mcla, previous_frontier_places)

    # Is the property reacheable ?
    reacheable = mcla.sq_is_satisfiable(query, steps)
    # If yes, in how many steps ?
    min_step = mcla.unfolder.current_step

    if reacheable and (min_step <= steps):
        LOGGER.info("%s:: Property %s is reacheable in %s steps",
                    final_prop, final_prop, min_step)
    else:
        LOGGER.info("%s:: Property %s is NOT reacheable in %s steps",
                    final_prop, final_prop, min_step)
        LOGGER.info("%s:: STOP the search!", final_prop)
        return

    # Find next MAC: Get FrontierSolution object
    frontier_sol = mcla.next_mac(query, min_step)
    if frontier_sol:
        LOGGER.debug("%s:: Next MAC object:\n%s", final_prop, frontier_sol)

        # Save MAC and timings
        LOGGER.debug("%s:: Save MAC and timings...", final_prop)
        with open(mac_complete_file, "a") as file:
            frontier_sol.save(file)

        # Save MAC (in alphabetic order...)
        LOGGER.debug("%s:: Save next MAC: %s", final_prop)
        with open(mac_file, "a") as file:
            frontier_sol.save(file, only_macs=True)

        # Save min steps
        min_step = mcla.unfolder.current_step - 1  # Magic number !
        # LOGGER.debug("%s:: Save minimal steps: %s", final_prop, min_step)
        # with open(mac_step_file, 'a') as file:
        #     file.write(str(min_step)+'\n')

        return frontier_sol.activated_frontier, min_step

    LOGGER.info("%s:: STOP the search! No more MAC.", final_prop)


def detect_model_type(mclanalyser, filepath):
    """Return the function to use to load the model.

    The detection is based on the file extension.

    bcx file: Build an MCLAnalyser from a .bcx file:
        build_from_chart_file()
    cal file: Build an MCLAnalyser from a .cal file of PID database
        build_from_cadlang()
    xml file: Build an MCLAnalyser from a .xml file of PID database:
        build_from_pid_file()

    :param arg1: MCLAnalyser.
    :param arg2: File that contains the model.
    :type arg1: <MCLAnalyser>
    :type arg2: <str>
    :return: The function to use to read the given file.
    :rtype: <func>
    """

    build_func = {
        ".bcx": mclanalyser.build_from_chart_file,
        ".cal": mclanalyser.build_from_cadlang,
        ".xml": mclanalyser.build_from_pid_file,
    }

    _, extension = os.path.splitext(filepath)
    LOGGER.debug("Found %s extension: %s", extension, filepath)

    if extension not in build_func:
        LOGGER.error("Unauthorized file: %s", filepath)
        exit()

    return build_func[extension]


# def main(model_file, mac_file, mac_step_file, mac_complete_file, mac_strong_file,
#         steps, final_prop, start_prop, inv_prop, all_macs):
#
#    LOGGER.debug("Params: start: {}, inv: {}, final: {}".format(start_prop,
#                                                                inv_prop,
#                                                                final_prop))
#
#    mac_p = None
#    # forbid previous mac
#    try :
#        mac_p = camFile2notOr(mac_file)
#        print("mac file:", mac_p)
#    except :
#        print('error in camFile2notOr')
#
#    if start_prop and mac_p :
#        start_prop += ' and ('+mac_p+')'
#    elif mac_p :
#        start_prop = mac_p
#
#    # BUILD MCLA
#    error_reporter = ErrorRep()
#    mcla = MCLAnalyser(error_reporter)
#    mcla.build_from_chart_file(model_file)
#
#    if error_reporter.error:
#        #cancel_warn(error_reporter.memory)
#        #error_reporter.reset()
#        #return
#        print("ERROR:: " + error_reporter.error)
#        raise
##
#    # BUILD QUERY
#    query = MCLSimpleQuery(start_prop, inv_prop, final_prop)
#
#    # Frontier init:
#    # on_yn function:
#    # reach = self.mcla.sq_is_satisfiable(query, max_step)
#
#    # Show solutions:
#    #on_cond function:
##    lsol = mcla.sq_frontier_solutions(query, step, 10)
##    print("DEBUG:: " + str(lsol))
##    print("DEBUG:: ", len( lsol))
#
#    # minimal activation conditions
#    # on_mac function:
##    mac_list = self.mcla.mac_search(query, self.max_step)
##            if len(mac_list)==0 :
##                ok_warn("The solver returns an empty list" +
##                        "\n"+" you should refine your query")
#
#    # OPTIMIZE STEP RESEARCH
#    if os.path.isfile(mac_step_file):
#        min_step = int(get_last_line(mac_step_file))
#        print("min_step opti:", min_step)
#        query.steps_before_check = min_step - 1
#
#
#    reacheable = mcla.sq_is_satisfiable(query, steps) #important step
#    print("reacheable:", reacheable)
#    min_step = mcla.unfolder.current_step
#    print("min_step:", min_step)
#    # Set max step authorized
#    query.steps_before_check = min_step - 1
#
#    # FIND NEXT MAC
#    next_mac_object = mcla.next_mac(query, min_step) #important step
#    print("next_mac_object:", next_mac_object)
#    if next_mac_object:
#
#        # SAVE MAC AND TIMING
#        with open(mac_complete_file, 'a') as file:
#            next_mac_object.save(file)
#
#        # SAVE MAC
#        next_mac = next_mac_object.activated_frontier
#        print("save next mac:", next_mac)
#        write_list(next_mac, mac_file)
#
#        # SAVE STEP
#        min_step = mcla.unfolder.current_step
#        print("save min step:", min_step)
#        with open(mac_step_file, 'a') as file:
#            file.write(str(min_step)+'\n')
#
#        return 0
#    else:
#        print("stop")
#        return 1


def compute_combinations(final_properties):
    """Return all combinations of final properties.

    .. note:: (in case of input_file and combinations set).

    :param: List of final properties.
    :type: <list>
    :return: List of str. Each str is a combination of final_properties linked
        by a logical 'and'.

        :Example:

            ``('TGFB1', 'COL1A1'), ('TGFB1', 'decorin')``
            gives: ``['TGFB1 and COL1A1', 'TGFB1 and decorin']``

    :rtype: <list <str>>
    """

    final_properties = set(final_properties)

    def get_formula(data):
        """Include all elements in the combinations and exclude explicitely,
        all other elements.
        """

        negated_places = " and not ".join(final_properties - data)
        return "{}{}{}".format(
            " and ".join(data),
            " and not " if negated_places != "" else "",
            negated_places,
        )

    all_combinations = list()
    for i in range(1, len(final_properties) + 1):

        all_combinations.append(
            {get_formula(set(comb)) for comb in it.combinations(final_properties, i)}
        )

    # Unpack combinations
    all_combinations = [comb for comb in it.chain(*all_combinations)]
    LOGGER.debug("Combinations: %s, Length: %s", all_combinations, all_combinations)
    LOGGER.info("Number of computed combinations: %s", len(all_combinations))

    return all_combinations


def solutions_search(params):
    """Launch the search for Minimum Activation Conditions (MAC) for entities
    of interest.

    * If there is no input file, there will be only one process.
    * If an input file is given, there will be 1 process per line
      (per logical formula on each line).
    """
    # No input file
    if params["final_prop"]:
        compute_macs(params)

    else:
        # Multiple properties in input file
        # => multiprocessing: 1 process for each property

        with open(params["input_file"], "r") as f_d:
            g = (line.rstrip("\n") for line in f_d)

            final_properties = [prop for prop in g if prop != ""]

        if params["combinations"]:
            # If input_file is set, we can compute all combinations of
            # final_properties. default: False
            final_properties = compute_combinations(final_properties)

        LOGGER.debug("Final properties: %s", final_properties)
        # Output combinations of final_properties
#        with open(params['input_file'] + '_combinations.txt', 'w') as f_d:
#            f_d.write('\n'.join(final_properties) + '\n')
#
#        g = (elem for elem in final_properties)
#        for i in range(1, len(final_properties) + 1):
#            with open(params['input_file'][:-4] + '_combinations' + str(i) + '.txt', 'w') as f_d:
#                try:
#                    f_d.write(next(g) + '\n')
#                    f_d.write(next(g) + '\n')
#                except StopIteration:
#                    break
#
#        exit()

        def update_params(prop):
            """Shallow copy of parameters and update final_prop for a new run"""
            new_params = params.copy()
            new_params["final_prop"] = prop
            return new_params

        # Fix number of processes
        # PS: the new solver is optimized for 8 threads
#        nb_cpu_required = mp.cpu_count() / 8
#        nb_cpu_required = 1 if nb_cpu_required == 0 else nb_cpu_required

        with ProcessPoolExecutor(max_workers=mp.cpu_count()) as executor:

            futures_and_output = {
                executor.submit(compute_macs, update_params(job_property)): job_property
                for job_property in final_properties
            }  # Job name

        nb_errors = 0
        nb_done = 0
        for future in as_completed(futures_and_output):

            job_name = futures_and_output[future]

            # On affiche les résultats si les futures en contiennent.
            # Si elles contiennent une exception, on affiche l'exception.
            if future.exception() is not None:
                LOGGER.error(
                    "%s generated an exception: \n%s", job_name, future.exception()
                )
                nb_errors += 1
            else:
                # The end
                LOGGER.info("%s... \t\t[Done]", job_name)
                nb_done += 1

        LOGGER.info("Ending: %s errors, %s done\nbye.", nb_errors, nb_done)


def compute_macs(params):
    """Launch Cadbiom search of MACs (Minimal Activation Conditions).

    This function is called 1 or multiple times according to the necessity
    to use multiprocessing (Cf launch_researchs()).

    .. note:: Previous result files will be deleted.

    """

    # Limit recursion
    sys.setrecursionlimit(10000)

    # QUERY PARAMETERS
    model_filename = os.path.basename(os.path.splitext(params["model_file"])[0])

    # FILES
    mac_file_prefix = params["output"] + model_filename + "_" + params["final_prop"] + "_mac"
    # mac_file
    mac_file          = mac_file_prefix + ".txt"
    # mac_step_file
    mac_step_file     = mac_file_prefix + "_step.txt"
    # mac_complete_file
    mac_complete_file = mac_file_prefix + "_complete.txt"
    # mac_strong_file
    mac_strong_file   = mac_file_prefix + "_strongA.txt"

    def remove_file(file):
        """Reset files"""
        try:
            os.remove(file)
        except OSError:
            pass

    if not params["continue"]:
        # Reset previous working files
        # PS: the reload is done in search_entry_point() function
        remove_file(mac_file)
        remove_file(mac_step_file)
        remove_file(mac_complete_file)
        remove_file(mac_strong_file)

    # MAC research
    search_entry_point(
        params["model_file"],  # model_file
        mac_file,              # mac_file
        mac_step_file,         # mac_step_file
        mac_complete_file,     # mac_complete_file
        mac_strong_file,       # mac_strong_file
        params["steps"],
        params["final_prop"],
        params["start_prop"],
        params["inv_prop"],
        params["all_macs"],
        params["continue"],
        params["limit"],
    )


def read_mac_file(file):
    """Return a list a fontier places already found in mac file

    .. TODO: use tools/solutions.py functions to reload frontiers
        => put these functions into the library...

    .. note:: use make_logical_formula() to get the new start_prop of the run.

    :param: Mac file of a previous run
    :type: <str>
    :return: A set a frontier places.
    :rtype: <set>
    """

    with open(file, "r") as f_d:
        return {tuple(line.rstrip("\n").split(" ")) for line in f_d}



# -*- coding: utf-8 -*-
# Copyright (C) 2017  IRISA
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# The original code contained here was initially developed by:
#
#     Pierre Vignet.
#     IRISA
#     Dyliss team
#     IRISA Campus de Beaulieu
#     35042 RENNES Cedex, FRANCE
"""
This module groups functions directly related to the parsing and the management
of the files generated by the solver of Cadbiom.

Here we find high-level functions to parse or clean *mac* files, and extract
all their data to a JSON format, a data interchange format that is humanly
readable and useful in programming.

Generic functions
~~~~~~~~~~~~~~~~~
* :meth:`~cadbiom_cmd.tools.solutions.get_query_from_filename`

Handle \*mac_complete.txt files
-------------------------------
* :meth:`~cadbiom_cmd.tools.solutions.load_solutions`
* :meth:`~cadbiom_cmd.tools.solutions.convert_solutions_to_json`

Handle \*mac\* files
--------------------
* :meth:`~cadbiom_cmd.tools.solutions.get_solutions`
* :meth:`~cadbiom_cmd.tools.solutions.get_mac_lines`
* :meth:`~cadbiom_cmd.tools.solutions.get_all_macs`

"""
#from __future__ import unicode_literals
from __future__ import print_function

# Standard imports
import os
import glob
from collections import defaultdict, Counter

# Library imports
import cadbiom.commons as cm

LOGGER = cm.logger()


## Generic #####################################################################

def get_query_from_filename(model_file, solution_file):
    """Return the query string according to the given model and solution filenames

    :Example:

        .. code-block:: python

            >>> get_query_from_filename(
            ...     "/path/model.bcx",
            ...     "/another_path/model_ENTITY_and_not_ENTITY_mac_complete.txt"
            ... )
            "ENTITY_and_not_ENTITY"

    :param model_file: Path of a bcx model.
    :param solution_file: Path of a solution file (\*mac\* file).
    :type model_file: <str>
    :type solution_file: <str>
    """

    model_filename = os.path.basename(os.path.splitext(model_file)[0])
    solution_filename = os.path.basename(os.path.splitext(solution_file)[0])
    # Remove model_filename from solution_filename
    solution_filename = solution_filename.replace(model_filename + '_', '')

    # Remove suffix from solution_filename
    if '_mac_complete' in solution_filename:
        # remove _mac_complete
        return solution_filename[:-13]

    elif '_mac' in solution_filename:
        # remove _mac
        return solution_filename[:-4]

## Handle *mac_complete.txt files ##############################################

def load_solutions(file):
    """Open a file with many solution/MACs (\*mac_complete.txt files)
    and yield them.

    :Example:

        .. code-block:: python

            >>> solutions = load_solutions('./solution_mac_complete.txt')
            >>> print([solution for solution in solutions])
            ("Ax Bx", [['h2', 'h00'], ['h3'], ['h0', 'h1'], ['hlast']])

    :param: File name
    :type: <str>
    :return: A generator of tuples of "frontier places" and a list of
        events in each step.

        :Example:

            .. code-block:: python

                ("Ax Bx", [['h2', 'h00'], ['h3'], ['h0', 'h1'], ['hlast']])

    :rtype: <tuple <str>, <list>>
    """

    sol_steps = defaultdict(list)
    sol = ""
    with open(file, 'r') as f_d:
        for line in f_d:
            #LOGGER.debug("Load_solutions :: line: " + line)
            # Remove possible \t separator from first line (frontier solution)
            line = line.rstrip('\n').rstrip('\t').replace('\t', ' ')
            # TODO: remove last space ' ' => beware, may be informative...
            # (start transitions have no event names: ori="__start__0")
            line = line.rstrip(' ')
            if line == '' or line[0] == '=':
                # Blank line
                # Skip blank lines and solution separator in some files (=====)
                continue
            elif line[0] != '%':
                if sol == line:
                    # TODO: why this condition ?
                    # => multiple lines with the same solution ?
                    # Same frontier places
                    yield sol, sol_steps[sol]

                    # reinit sol
                    sol_steps[sol] = list()
                    continue
                elif sol == '':
                    # First sol
                    sol = line
                else:
                    # Yield previous sol
                    yield sol, sol_steps[sol]
                    sol = line

            elif line[0] == '%':
                # Remove step with only "% "
                step = line.lstrip('% ')

                if step != '':
                    sol_steps[sol].append(step.split(' '))

        # Yield last sol
        yield sol, sol_steps[sol]


def convert_solutions_to_json(sol_steps, transitions, conditions=True):
    """Convert all events for all solutions in a complete MAC file
    and write them in a separate file in the JSON format.

    This is a function to quickly search all transition attributes involved
    in a solution.

    :Example:

        .. code-block:: python

            >>> from tools.models import get_transitions
            >>> # Get transitions from the model
            >>> model_transitions = get_transitions('model.bcx')
            >>> decomp_solutions = convert_solutions_to_json(
            ...     load_solutions('./solution_mac_complete.txt'),
            ...     model_transitions,
            ...     conditions=True,
            ... )
            >>> print(decomp_solutions)
            [{
                "solution": "Ax Bx",
                "steps": [
                    [{
                        "event": "_h_2",
                        "transitions": [{
                            "ext": "n3",
                            "ori": "Bx"
                        }]
                    }],
                ]
            }]


    :param arg1: List of steps involved in a solution. See load_solutions().
        A tuple of "frontier places" and a list of events in each step.
        ``("Bx Ax", [['h2', 'h00'], ['h3'], ['h0', 'h1'], ['hlast']])``
    :param arg2: A dictionnary of events as keys, and transitions as values.
        Since many transitions can define an event, values are lists.
        Each transition is a tuple with: origin node, final node, attributes
        like label and condition.
        ``{'h00': [('Ax', 'n1', {'label': 'h00[]'}),]``
        See get_transitions().
    :param arg3: (Optional) Integrate in the final file,
        the conditions for each transition.
    :type arg1: <list>
    :type arg2: <dict <list <tuple <str>, <str>, <dict <str>: <str>>>>
    :type arg3: <bool>
    :return: Return the JSON data for the given steps.

        Example:

        .. code-block:: javascript

            [{
                "solution": "Ax Bx",
                "steps": [
                    [{
                        "event": "_h_2",
                        "transitions": [{
                            "ext": "n3",
                            "ori": "Bx"
                        }]
                    }],
                ]
            }]
    :rtype: <list>
    """

    def get_transition_def(step_event):
        """Dump each transition in the given event to a list of dictionnaries.

        .. note:: ori="JUN_nucl_gene" ext="JUN_nucl" event="_h_391"
        :return: A list of dictionaries
            (1 dict for 1 transition in the given event)
        :rtype: <list <dict>>
        """

        # Many transitions per event (ex: complex dissociation)
        decomp_transitions = list()
        for trans in step_event:

            decomp_transition = {
                #"event": trans[2]['label'].split('[')[0],
                "ori": trans[0],
                "ext": trans[1],
            }
            # If condition boolean is set (by default),
            # we add the event's transition to the json data.
            if conditions:
                decomp_transition["condition"] = trans[2]['condition']

            decomp_transitions.append(decomp_transition)

        return decomp_transitions


    # sol_steps structure:
    # ("Bx Ax", [[u'h2', u'h00'], [u'h3'], [u'h0', u'h1'], [u'hlast']])
    decomp_solutions = list()
    for sol, steps in sol_steps:
        # Decompile steps in each solution
        decomp_steps = list()
        for step in steps:
            # Decompile events in each step
            decomp_events = list()
            for event in step:
                # Decompile transitions in each event
                decomp_event = dict()
                # Get transitions for the given event
                # Structure of transitions:
                # {u'h00': [('Ax', 'n1', {u'label': u'h00[]'}),]
                step_event = transitions.get(event, None)
                if not step_event:
                    LOGGER.error(
                        "convert_solutions_to_json:: event not found in the "
                        "transitions of the model: %s",
                        event
                    )
                    LOGGER.error(
                        "convert_solutions_to_json:: transitions: %s",
                        transitions
                    )
                    decomp_event['event'] = "ERROR, no transition"
                else:
                    # Get list of transitions
                    decomp_event['event'] = event
                    decomp_event['transitions'] = get_transition_def(step_event)
                # Add event and its transitions
                decomp_events.append(decomp_event)
            # Add step and its events
            decomp_steps.append(decomp_events)
        # Add solution and its steps
        solution = {
            "solution": sol,
            "steps": decomp_steps,
        }
        decomp_solutions.append(solution)

    return decomp_solutions

## Handle *mac* files ##########################################################

def get_solutions(file_descriptor):
    """Generator of solution lines and corresponding stripped lines for
    \*mac\* file.

    .. note:: This function does not return events! It is just original lines and
        cleaned lines containing solutions (i.e sets of frontier
        places/boundaries).

        We remove the last ``'\\n'`` and ``'\\t'``.
        Tabs in the middle are replaced by one space ``' '``.

    :param: Opened file.
    :type: <file>
    :return: A generator of tuples; each tuple contains the original line,
        and the cleaned line.

        :Example:

            For an original line: ``'Z\\tY\\tX\\n'``

            .. code-block:: python

                ('Z\\tY\\tX', 'X Y Z')

    :rtype: <tuple <str>, <str>>
    """

    for line in file_descriptor:
        # Remove possible \t separator from first line (frontier solution)
        line = line.rstrip('\n')
        stripped_line = line.rstrip('\t').replace('\t', ' ')

        # Next Line if empty
        if stripped_line == '':
            continue

        # Remove events or other lines
        if stripped_line[0] not in ('%', '=', ' '):
            # print(stripped_line)
            # Sort in lower case, remove ' ' empty elements
            yield line, stripped_line


def get_mac_lines(filepath):
    """Returns only a set of MAC LINES from A file.

    This function is based on :meth:`~cadbiom_cmd.tools.solutions.get_solutions`
    that returns mac lines and stripped mac lines.

    .. note:: You would prefer to use :meth:`~cadbiom_cmd.tools.solutions.get_all_macs`
        which:

        - Can handle a directory path and return all macs in it,
        - Can handle a simple file,
        - Do some verifications on all parsed macs.

    .. note:: We assume that at this point, all MAC lines are sorted in
        alphabetical order.

    .. note:: We return LINES not a set of places.

        :Example:

            .. code-block:: python

                    {'Cx Dx', 'Ax Bx'}

    :param: Filepath to be opened and in which solutions will be returned.
    :type: <str>
    :return: Set of MAC/CAM from the given file.
    :rtype: <set <str>>
    """

    # Return mac lines
    with open(filepath, 'r') as f_d:
        return {stripped_line for _, stripped_line in get_solutions(f_d)}


def get_all_macs(path):
    """Return a set of all MAC LINES from a directory or from a file.

    This function is based on :meth:`~cadbiom_cmd.tools.solutions.get_solutions`
    that returns mac lines and stripped mac lines, and
    :meth:`~cadbiom_cmd.tools.solutions.get_mac_lines` that returns only mac
    lines from a file.


    .. note:: Alternatively we do some verifications here:

        - Detection of duplicated MACS (AssertionError raised)
        - Print number of MACS per file
        - Print duplicated MACS
        - Print number of MACS

    :param: Filepath to be opened and in which solutions will be returned.
    :type: <str>
    :return: Set of MAC/CAM from the given path.
    :rtype: <set <str>>
    """

    # Put all macs in a list, not a set
    # => allow us to detect duplicated macs among all mac files
    total_macs = list()
    file_number = 0
    if os.path.isfile(path):
        # Get set of mac lines
        total_macs = get_mac_lines(path)
        file_number = 1

    elif os.path.isdir(path):

        # Get set of macs for each file
        for file_number, file in enumerate(glob.glob(path + '*mac.txt'), 1):

            temp_macs = get_mac_lines(file)
            total_macs += list(temp_macs)

            # Print the number of macs for the given file
            LOGGER.info("File %s: MACS: %s", file, len(temp_macs))

    assert file_number != 0, "No *mac.txt files found!"
    LOGGER.info("Files processed: %s", file_number)

    # Check of duplicated macs (number > 1)
    unique_macs = set(total_macs)
    duplicated_macs = {(k, v) for k, v in Counter(total_macs).items() if v != 1}

    LOGGER.info("Duplicated MACS: %s", duplicated_macs)
    LOGGER.info("Number of MACS loaded: %s", len(total_macs))
    LOGGER.info("Number of unique MACS returned: %s", len(unique_macs))

    # assert len(total_macs) == len(unique_macs)
    return unique_macs

## Handle *.json files #########################################################




# -*- coding: utf-8 -*-
# Copyright (C) 2017  IRISA
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# The original code contained here was initially developed by:
#
#     Pierre Vignet.
#     IRISA
#     Dyliss team
#     IRISA Campus de Beaulieu
#     35042 RENNES Cedex, FRANCE
"""Handle generated files

This module provides some functions to do some analyzis on the output
files of Cadbiom.

Entry points:

    * :meth:`~cadbiom_cmd.solution_sort.queries_2_json`
    * :meth:`~cadbiom_cmd.solution_sort.solutions_2_graphs`
    * :meth:`~cadbiom_cmd.solution_sort.solutions_2_common_graph`

:Example of the content of a complete solution file:

    .. code::

        Bx  Ax
        % h2 h00
        % h3
        % h0 h1
        % hlast
        Bx  Ax
        % h2
        % h3 h00
        % h0 h1
        %
        % hlast
        Bx  Ax
        % h2
        % h3 h00
        % h0 h1
        % hlast
        %
        %
        Bx  Ax
        % h2 h00
        % h3
        % h0 h1
        % hlast
        %
        %
        %

"""
from __future__ import unicode_literals
from __future__ import print_function

# Standard imports
from collections import defaultdict, Counter
import itertools as it
import json
import os
import glob
import csv

# Library imports

from tools.solutions import get_solutions
from tools.models import get_transitions_from_model_file
from tools.solutions import load_solutions, convert_solutions_to_json, \
    get_query_from_filename, get_mac_lines
from tools.graphs import export_graph, build_graph, merge_graphs, get_solutions_graph_data

import cadbiom.commons as cm

LOGGER = cm.logger()

## Sort functions ##############################################################

def sort_solutions_in_file(filepath):
    """Sort all solutions in the given file in alphabetical order.

    .. warning:: The file is modified in place.

    :param: Filepath to be opened and in which solutions will be sorted.
    :arg: <str>
    """

    solutions = dict()

    with open(filepath, 'r+') as f_d:

        # Get old line as key and ordered line as value
        for line, stripped_line in get_solutions(f_d):
            # Sort in lower case, remove ' ' empty elements
            solutions[line] = \
                " ".join(sorted([place for place in stripped_line.split(' ')
                                 if place != ' '], key=lambda s: s.lower()))

        # Rewind the whole file
        f_d.seek(0)

        # Load all the content
        file_text = f_d.read()

        # Replace old sols with the new ones
        for original_sol, sorted_sol in solutions.items():
            file_text = file_text.replace(original_sol, sorted_sol)

        # Rewind the whole file
        f_d.seek(0)

        # Write all text in the current opened file
        f_d.write(file_text)


def solutions_sort(path):
    """Entry point for sorting solutions.

    Read a solution(s) file(s) (\*mac\* files) and sort all
    frontier places/boundaries in alphabetical order.

    This functions tests if the given path is a directory or a file.

    .. warning:: The files will be modified in place.

    :param: Filepath or directory path containing Cadbiom solutions.
    :type: <str>
    """

    # Check valid input file/directory
    assert os.path.isfile(path) or os.path.isdir(path)

    if os.path.isdir(path):
        # Recursive search of *mac* files
        # (mac.txt, mac_complete.txt, mac_step.txt)
        path = path if path[-1] == '/' else path + '/'
        [sort_solutions_in_file(file) for file in glob.glob(path + '*mac*')]
    else:
        sort_solutions_in_file(path)

## Conversion functions ########################################################

def write_json(output_dir, file_path, file_suffix, data):
    """Write decompiled solutions to a JSON formated file

    Called by :meth:`queries_2_json TODO` and  :meth:`queries_2_common_graph`

    :param output_dir: Output directory
    :param file_path: Filepath of the original solution file.
        We extract the basename in order to name the JSON file.
    :param file_suffix: String added to the solution filename.
        Ex: filename + file_suffix + ".json"
    :param data: Data to be serialized in JSON
    :type output_dir: <str>
    :type file_path: <str>
    :type file_suffix: <str>
    :type data: <list> or <dict> or <whatever>
    """
    # Add file_suffix to the solution filename
    filename = os.path.basename(os.path.splitext(file_path)[0])

    with open(output_dir + filename + file_suffix + '.json', 'w') as f_d:
        json.dump(data, f_d, sort_keys=True, indent=2)

## query - json
def queries_2_json(output_dir, model_file, path, conditions=True):
    """Entry point for queries_2_json

    Create a JSON formated file containing all data from complete MAC files
    (\*mac_complete files). The file will contain frontier places/boundaries
    and decompiled steps with their respective events for each solution.

    This is a function to quickly search all transition attributes involved
    in a solution.

    This functions tests if the given path is a directory or a file.

    :param output_dir: Output path.
    :param model_file: Filepath of the model.
    :param path: Filepath/directory of a complete solution file.
    :param conditions: (Optional) If False, conditions of transitions will not
        be present in the JSON file. This allows to have only places/entities
        used inside trajectories; thus, inhibitors are avoided.
    :type output_dir: <str>
    :type model_file: <str>
    :type path: <str>
    :type conditions: <boolean>
    """
    # Check valid input file/directory
    assert os.path.isfile(path) or os.path.isdir(path)

    # Get transitions from the model
    model_transitions, _ = get_transitions_from_model_file(model_file)

    if os.path.isfile(path):
        # The given path is a solution file
        decomp_solutions = convert_solutions_to_json(
            load_solutions(path),
            model_transitions,
            conditions=conditions,
        )
        write_json(output_dir, path, "_decomp", decomp_solutions)

    elif os.path.isdir(path):
        # The given path is a directory
        path = path if path[-1] == '/' else path + '/'

        # Decompilation of all files in the directory
        file_number = 0
        for file_number, solution_file in \
            enumerate(glob.glob(path + '*mac_complete.txt'), 1):

            decomp_solutions = convert_solutions_to_json(
                load_solutions(solution_file),
                model_transitions,
                conditions=conditions,
            )
            write_json(solution_file, path, "_decomp", decomp_solutions)

        LOGGER.info("Files processed: %s", file_number)
        assert file_number != 0, "No *mac_complete.txt files found!"

################################################################################

def solutions_2_graphs(output_dir, model_file, path):
    """Entry point for solutions_2_graphs

    Create GraphML formated files containing a representation of the
    trajectories for each solution in complete MAC files (\*mac_complete files).

    This is a function to visualize paths taken by the solver from the boundaries
    to the entities of interest.

    This functions tests if the given path is a directory or a file.

    :param output_dir: Output path.
    :param model_file: Filepath of the model.
    :param path: Filepath/directory of a/many complete solutions files.
    :type output_dir: <str>
    :type model_file: <str>
    :type path: <str>
    """

    # Check valid input file/directory
    assert os.path.isfile(path) or os.path.isdir(path)

    # Get transitions from the model
    model_transitions, _ = get_transitions_from_model_file(model_file)

    if os.path.isfile(path):
        # The given path is a solution file
        save_solutions_to_graphs(
            output_dir,
            load_solutions(path),
            model_transitions
        )

    elif os.path.isdir(path):
        # The given path is a directory
        path = path if path[-1] == '/' else path + '/'

        # Decompilation of all files in the directory
        file_number = 0
        for file_number, solution_file in \
            enumerate(glob.glob(path + '*mac_complete.txt'), 1):

            save_solutions_to_graphs(
                output_dir,
                load_solutions(solution_file),
                model_transitions
            )

        LOGGER.info("Files processed: %s", file_number)
        assert file_number != 0, "No *mac_complete.txt files found!"


def save_solutions_to_graphs(output_dir, sol_steps, transitions):
    """Build and export graphs based on the given solutions

    Each solution is composed of a set of frontier places and steps,
    themselves composed of events.
    We construct a graph based on the transitions that occur in the composition
    of the events of the given solution.

    :param output_dir: Output path.
    :param sol_steps: A generator of tuples of "frontier places" and a list of
        events in each step.

        :Example:

            .. code-block:: python

                ("Bx Ax", [['h2', 'h00'], ['h3'], ['h0', 'h1'], ['hlast']])

    :param transitions: A dictionnary of events as keys, and transitions as values.
        Since many transitions can define an event, values are lists.
        Each transition is a tuple with: origin node, final node, attributes
        like label and condition.

        :Example:

            .. code-block:: python

                {'h00': [('Ax', 'n1', {'label': 'h00[]'}),]

    :type output_dir: <str>
    :type sol_steps: <tuple <str>, <list>>
    :type transitions: <dict <list <tuple <str>, <str>, <dict <str>: <str>>>>
    """

    for sol_index, (sol, steps) in enumerate(sol_steps):

        # build_graph() returns :
        # G, transition_nodes, all_nodes, edges_in_cond, edges
        # sol_index is used to order files according to the order of appearance
        # in the file
        export_graph(output_dir, sol, sol_index,
                     build_graph(sol, steps, transitions)[0])

################################################################################

def get_solution_graphs(sol_steps, transitions):
    """Generator that yields the graphs of the given solutions.

    .. note:: See the doc of a similar function
        :meth:`~cadbiom_cmd.solution_sort.save_solutions_to_graphs`.
    """
    for sol, steps in sol_steps:
        # build_graph() returns :
        # G, transition_nodes, all_nodes, edges_in_cond, edges
        # Python 3: partial unpacking: G, *_
        yield build_graph(sol, steps, transitions)[0]


def queries_2_common_graph(output_dir, model_file, path,
                           make_graphs=True, make_csv=False, make_json=False,
                           *args, **kwargs):
    """Entry point for solutions_2_common_graph

    Create a GraphML formated file containing a unique representation of **all**
    trajectories corresponding to all solutions in each complete MAC files
    (\*mac_complete files).

    This is a function to visualize paths taken by the solver from the boundaries
    to the entities of interest.

    CSV fields::

        - query
        - solutions: nb trajectories/solutions
        - boundaries
        - events
        - genes
        - Protein
        - Complex
        - influencing_places
        - graph_nodes
        - graph_nodes_places
        - graph_edges
        - strongly_connected
        - max_degree
        - min_degree
        - average_degree

    .. note:: This functions tests if the given path is a directory or a file.

    :param output_dir: Output path.
    :param model_file: Filepath of the model.
    :param path: Filepath/directory of a/many complete solutions files.
    :key make_graphs: (optional) Make a GraphML for each query results in path.
        default: True
    :key make_csv: (optional) Make a **global** CSV for all query results in path.
        default: False
    :key make_json: (optional) Make a JSON dump of each query results in path.
        default: False
    :type output_dir: <str>
    :type model_file: <str>
    :type path: <str>
    :type make_graphs: <boolean>
    :type make_csv: <boolean>
    :type make_json: <boolean>
    """

    def write_csv(data):
        """Write given data in CSV file; then flush the file descriptor

        .. note:: data is modified in place.

        :param data: Data to be serialized in JSON.
        """
        # Handle nested dictionaries by flattening them
        data.update({
            sub_key: sub_val
            for key, val in data.items() if isinstance(val, dict)
            for sub_key, sub_val in val.items()
        })

        writer.writerow(data)
        f_d.flush()


    def do_magic(solution_file):
        # Get query string from the name of the solution file
        query = get_query_from_filename(model_file, solution_file)

        LOGGER.info("Processing %s query...", query)

        # generator of ("Ax Bx", [['h2', 'h00'], ['h3'], ['h0', 'h1'], ['hlast']])
        solutions = tuple(load_solutions(solution_file))
        graphs = get_solution_graphs(
            solutions,
            model_transitions
        )
        # Get common graph
        graph = merge_graphs(graphs)

        if make_graphs:
            # Write graph
            export_graph(output_dir, query, '', graph)

        if make_json or make_csv:
            # Export to json file (similaire à model_graph_info() ...)

            # Get 2 lists: 1: frontiers; 2: events
            unzipped_solutions = list(zip(*solutions))
            # Frontiers
            frontiers = set(it.chain(*[sol.split() for sol in unzipped_solutions[0]]))
            # Events
            events = set(it.chain(*it.chain(*unzipped_solutions[1])))
            # Genes
            # brutal but effective as long as the nodes of a model do not have
            # the information in hard in one of their attributes
            genes = {name for name in frontiers if "_gene" in name}

            # Influencing frontier places
            g = (trans.get_influencing_places() for trans in parser.model.transition_list if trans.event in events)
            influencing_places = frontiers & set(it.chain(*g))

            # Get frontier types
            # Do not count genes entities that are "virtual" types
            g = (parser.model.node_dict[name] for name in frontiers - genes)
            # Remove nodes with empty notes
            frontier_types = Counter([
                json.loads(node.note).get("entityType", "UKN") for node in g
                if node.note
            ])

            solutions_info = {
                'modelFile': model_file,
                'query': query,
                'solutions': len(solutions),  # nb trajectories/solutions
                'boundaries': len(frontiers),
                'events': len(events),
                'genes': len(genes),
                'influencing_places': len(influencing_places),
                'entityTypes': frontier_types,
            }
            # Advanced metrics
            get_solutions_graph_data(graph, solutions_info, True)
            LOGGER.debug("%s", solutions_info)

            if make_json:
                # Save to <solution_filename>_graph_summary.json
                write_json(output_dir, solution_file, "_graph_summary", solutions_info)

            if make_csv:
                # Export to csv graphs_summary.csv
                write_csv(solutions_info)


    # Check valid input file/directory
    assert os.path.isfile(path) or os.path.isdir(path)

    # Bind arguments to avoid overwriting previous imports
    make_graphs = kwargs['graphs']
    make_json = kwargs['json']
    make_csv = kwargs['csv']

    # Get transitions from the model
    model_transitions, parser = get_transitions_from_model_file(model_file)


    if make_csv:
        # Init the CSV file now
        f_d = open(output_dir + 'graphs_summary.csv', 'w')
        # Get a subset of solutions_info keys as fieldnames
        writer = csv.DictWriter(
            f_d,
            [
                "query", "solutions", "boundaries", "events", "genes",
                "Protein", "Complex", "influencing_places",
                "graph_nodes", "graph_nodes_places", "graph_edges",
                "strongly_connected",
                "max_degree", "min_degree", "average_degree",
            ],
            extrasaction="ignore",  # Ignore unknown fieldnames
            restval=0,  # Value for missing fieldnames
            delimiter=str(';'),
        )
        writer.writeheader()


    # File management...
    if os.path.isfile(path):
        # The given path is a solution file
        do_magic(path)

    elif os.path.isdir(path):
        # The given path is a directory
        path = path if path[-1] == '/' else path + '/'

        # Decompilation of all files in the directory
        file_number = 0
        for file_number, solution_file in \
                enumerate(glob.glob(path + '*mac_complete.txt'), 1):

            do_magic(solution_file)

        LOGGER.info("Files processed: %s", file_number)
        assert file_number != 0, "No *mac_complete.txt files found!"


    if make_csv:
        # Close the CSV file descriptor
        f_d.close()

## Matrices of occurrences #####################################################

def queries_2_occcurrence_matrix(output_dir, model_file, path,
                                    transposed=False, normalized=False):
    """Entry point for queries_2_occcurrence_matrix

    See :meth:`~cadbiom_cmd.solution_sort.occurrence_matrix`.

    :param output_dir: Output path.
    :param model_file: Filepath of the model.
    :param path: Directory of many complete solutions files.
    :param transposed: (Optional) Transpose the final matrix (switch columns and rows).
    :type output_dir: <str>
    :type model_file: <str>
    :type path: <str>
    :type transposed: <boolean>
    """

    # Check valid input directory
    assert os.path.isdir(path)

    path = path if path[-1] == '/' else path + '/'

    # Make matrix
    occurrence_matrix(output_dir, model_file, path)

    if transposed:
        transpose_csv(input_file=output_dir + 'occurrence_matrix.csv',
                      output_file=output_dir + 'occurrence_matrix_t.csv')


def occurrence_matrix(output_dir, model_file, path,
                     matrix_filename='occurrence_matrix.csv'):
    """Make a matrix of occurrences for the solutions in the given path.

    - Compute occurrences of each place in all `mac.txt` files.
    - Save the matrix in csv format with the following columns:
        Fieldnames: "patterns (number)/places (number);mac_number;frontier places"
        Each request (pattern) is accompanied by the number of solutions found.

    .. todo:: Split the creation and writing of the matrix in 2 functions.

    :param output_dir: Output path.
    :param model_file: Filepath of the model.
    :param path: Directory of many complete solutions files.
    :param matrix_filename: (Optional) Filename of the matrix file.
    :type output_dir: <str>
    :type model_file: <str>
    :type path: <str>
    :type matrix_filename: <str>
    :return: A dictionnary with the matrix object.
        keys: queries, values: occurrences of frontier places
    :rtype: <dict>
    """

    # Key: Logical formula as input of Cadbiom
    # Value: Number of each place in all solutions of the current file
    matrix = defaultdict(Counter)
    # All frontier places in all mac files
    all_frontier_places = set()

    # Compute occurrences of each place in all mac files
    file_number = 0
    for file_number, filepath in enumerate(glob.glob(path + '*mac.txt'), 1):

        # gene pattern
#        pattern = {gene for gene in genes if gene in mac}

        # Get query string from the name of the solution file
        # From: 'MODEL_NAME_PLACE1 and not PLACE2 and not PLACE3_mac.txt'
        # Get: 'PLACE1 and not PLACE2 and not PLACE3'
        query = get_query_from_filename(model_file, filepath)

        mac_number = 0
        for mac_number, mac_line in enumerate(get_mac_lines(filepath), 1):

            frontier_places = set(mac_line.split(' '))
            # Update set of all frontier places
            all_frontier_places.update(frontier_places)
            # Update counter of places => compute frequencies
            matrix[query] += Counter(frontier_places)

        # Set the mac_number for future standardization
        matrix[query]["mac_number"] = mac_number

    LOGGER.info("Files processed: %s", file_number)
    assert file_number != 0, "No *mac.txt files found!"

    # Save the matrix
    # columns: "patterns (number)/places (number);mac_number;frontier places"
    with open(output_dir + matrix_filename, 'w') as f_d:

        # Forge header
        header = "patterns ({})/places ({})".format(
            len(matrix),
            len(all_frontier_places),
        )
        writer = csv.DictWriter(
            f_d,
            delimiter=str(';'),
            restval=0, # default value for frequency
            fieldnames=[header, "mac_number"] + list(all_frontier_places))
        writer.writeheader()

        # Add a last line in the csv: total of occurrences for each place
        global_frontier_counter = Counter()
        # The first column is composed of the query + the number of solutions for it
        for query, row in matrix.iteritems():
            global_frontier_counter += row
            # PS: THIS modifies the matrix by adding a new key ('header')
            row[header] = "{} ({})".format(query, row["mac_number"])
            writer.writerow(row)

        # Total of occurrences at the end of the file
        global_frontier_counter[header] = "Total of occurrences"
        writer.writerow(global_frontier_counter)

    return matrix


def transpose_csv(input_file='occurrence_matrix.csv',
                  output_file='occurrence_matrix_t.csv'):
    """Useful function to transpose a csv file x,y => y,x

    .. note:: The csv file must be semicolon ';' separated.

    :param input_file: Input file.
    :param output_file: Output file transposed.
    :type input_file: <str>
    :type output_file: <str>
    """

    # Transpose file
    # PS: izip('ABCD', 'xy') --> Ax By
    data = it.izip(*csv.reader(open(input_file, "r"), delimiter=str(';')))
    csv.writer(open(output_file, "w"), delimiter=str(';')).writerows(data)



# -*- coding: utf-8 -*-
# Copyright (C) 2017  IRISA
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# The original code contained here was initially developed by:
#
#     Pierre Vignet.
#     IRISA
#     Dyliss team
#     IRISA Campus de Beaulieu
#     35042 RENNES Cedex, FRANCE
"""
This module groups functions directly related to the creation and the management
of the graph based on a Cadbiom model.

Here we find high-level functions to create a Networkx graph, and convert it
to JSON or GraphML formats.
"""
# Standard imports
import itertools as it
import datetime as dt
import networkx as nx

# Library imports
from cadbiom_cmd.tools.models import parse_condition

import cadbiom.commons as cm

LOGGER = cm.logger()


def build_graph(solution, steps, transitions):
    """Build a graph for the given solution.

        - Get & make all needed edges
        - Build graph

    .. note:: Legend:

        - Default nodes: grey
        - Frontier places: red
        - Transition nodes: blue
        - Inhibitors nodes: white
        - Default transition: grey
        - Inhibition edge: red
        - Activation edge: green

    :param arg1: Frontier places.
    :param arg2: List of steps (with events in each step).
    :param arg3: A dictionnary of events as keys, and transitions as values
        (see get_transitions()).
    :type arg1: <str>
    :type arg2: <list <list>>
    :type arg3: <dict <list <tuple <str>, <str>, <dict <str>: <str>>>>
    :return:
        - Networkx graph object.
        - Nodes corresponding to transitions with conditions.
        - All nodes in the model
        - Edges between transition node and nodes in condition
        - Normal transitions without condition
    :rtype: <networkx.classes.digraph.DiGraph>, <list>, <list>, <list>, <list>
    """

    def filter_transitions(step_event):
        """ Insert a transittion in a transition event if there is a condition.

        => Insert a node in a edge.
        => Link all nodes involved in the condition with this new node.

        :param: A list of events (transitions) (from a step in a solution).
            [('Ax', 'n1', {u'label': u'h00[]'}),]
        :type: <tuple>
        :return: Fill lists of edges:
            edges_with_cond: link to a transition node for
                transition with condition.
            transition_nodes: add new nodes corresponding to transitions with
                conditions.
            edges_in_cond: Add all edges to nodes linked to a transition
                via the condition (group of nodes involved in the path
                choosen by the solver).
            edges: transition untouched if there is no condition.
        :rtype: None
        """
        assert step_event # Todo: useful ?

        # Inactivated nodes in paths of conditions
        # This variable is set in place by parse_condition() for the current context
        inhibitors_nodes = set()

        input_places = {ori for ori, _, _ in step_event}

        # Color nodes
        # Since we explore all possible paths for each condition,
        # some edges are rewrited multiple times.
        # => included edges between origin <=> transition node
        # These edges must be grey while, edges between a node that is
        # only in a condition and a transition node must be green.
        # => notion of activator vs inhibitor vs normal input/output node
        def color_map(node):
            """Return a color used for edges, related to the role of the node"""
            # print("color for:", node)
            if node in inhibitors_nodes: # Test first (see cond below)
                return 'red'
            if node in input_places: # some/all frontier places are in this set
                return 'grey'

            return 'green'


        for trans in step_event:
            attributes = trans[2]
            ori = trans[0]
            ext = trans[1]
            event = attributes['label'].split('[')[0]

            # If there is a condition formula associated to this clock
            if attributes['condition'] != '':

                # Add the transition as node
                # PS: There may be a rewrite of an identical node here,
                # networkx will do the merge later, but the conditions
                # will not be preserved. Only the last condition is kept.
                transition_nodes.append(
                    (
                        event,
                        {
                            'name': attributes['label'], # Trick for Cytoscape
                            'color': 'blue',
                            'condition': attributes['condition'], # Partial cond (cf above)
                        }
                    )
                )

                # Origin => transition node
                edges_with_cond.append(
                    (
                        ori, event,
                        {
                            'label': ori + '-' + event,
                        }
                    )
                )

                # Transition node => ext
                edges_with_cond.append(
                    (
                        event, ext,
                        {
                            'label': event + '-' + ext,
                        }
                    )
                )

                # Add all transitions to nodes linked via the condition
                valid_paths = parse_condition(
                    attributes['condition'],
                    all_nodes,
                    inhibitors_nodes # Set in place by parse_condition
                )
                for i, path in enumerate(valid_paths):
                    for node in path:
                        edges_in_cond.append(
                            (
                                node, event,
                                {
                                    'label': '{} ({})'.format(
                                        event,
                                        i
                                    ),
                                    #'label': '{} [{}] ({})'.format(
                                    #    event,
                                    #    ', '.join(path),
                                    #    i
                                    #), #node + '-' + event,
                                    'color': color_map(node), # Set edge color
                                }
                            )
                        )
            else:
                # Normal edges
                edges.append(trans)


    # Get & make all needed edges ##############################################
    LOGGER.debug("Build graph for the solution: %s", solution)
    LOGGER.debug("Decompiled steps: %s", steps)

    frontier_places  = solution.split(' ')
    edges_with_cond  = list() # Edges between ori <-> transition node <-> ext
    edges_in_cond    = list() # Edges between transition node and nodes in condition
    transition_nodes = list() # Nodes inserted because of condition in transition
    edges            = list() # Normal transitions without condition

    # Get all nodes in all transitions (origin & ext)
    try:
        all_transitions = \
            (transitions[step_event] for step_event in it.chain(*steps))
        transitions_ori_ext = \
            (tuple((ori, ext)) for ori, ext, _ in it.chain(*all_transitions))
    except KeyError:
        LOGGER.error("/!\One event is not in the given model file... Please check it.")
        raise
    # All nodes are: nodes involved in transitions + frontier places
    all_nodes = set(it.chain(*transitions_ori_ext)) | set(frontier_places)

    # print("SOLUTION", solution)
    # print("ALL NODES", all_nodes)

    # Parse all conditions in transitions;
    # add nodes in conditions and transition nodes
    [filter_transitions(transitions[step_event])
     for step_event in it.chain(*steps)]

    # print("edges without cond", edges)
    # print("edges with cond", edges_with_cond)
    # print("transition nodes added", transition_nodes)

    # Make Graph ###############################################################
    G = nx.DiGraph()
    # Add all nodes (some frontier places are in this set)
    G.add_nodes_from(all_nodes, color='grey')
    # Add fontier places
    # We rewrite the color of these special nodes here.
    G.add_nodes_from(frontier_places, color='red')
    # Add all transition nodes
    G.add_nodes_from(transition_nodes, color='blue')

    # Node attribute ?
    # print(G.node['h1'])

    # Add all edges
    G.add_edges_from(edges)
    G.add_edges_from(edges_with_cond)
    G.add_edges_from(edges_in_cond)

    return G, transition_nodes, all_nodes, edges_in_cond, edges


def get_solutions_graph_data(G, info, centralities):
    """Complete the given dictionary with information specific to the graph considered

    Doc::

        https://networkx.github.io/documentation/networkx-1.10/reference/algorithms.component.html
        https://networkx.github.io/documentation/stable/reference/algorithms/shortest_paths.html
        average_shortest_path_length
        https://networkx.github.io/documentation/stable/reference/algorithms/generated/networkx.algorithms.shortest_paths.generic.average_shortest_path_length.html#networkx.algorithms.shortest_paths.generic.average_shortest_path_length
        weakly_connected_component_subgraphs
        https://networkx.github.io/documentation/networkx-1.10/reference/generated/networkx.algorithms.components.weakly_connected.weakly_connected_component_subgraphs.html#networkx.algorithms.components.weakly_connected.weakly_connected_component_subgraphs
        Measures
        https://networkx.github.io/documentation/stable/reference/algorithms/index.html

    By default the following information are added::

        - graph_nodes: Number of nodes
        - graph_edges: Number of edges
        - graph_nodes_places: Number of biological places/entities.
          The graph is a false bipartite graph, we remove the subset of transitions
          in order to have the real count of biological places/entities.

    If centralities is True, the folliwing information are added to the a new
    key named "centralities"::

        - strongly_connected:
        - weakly_connected
        - max_degree
        - min_degree
        - average_degree
        - degree
        - connected_components_number
        - connected_components
        - average_shortest_paths

    :param G: NetworkX directed graph
    :param info: Dictionnary of data to be completed
    :param centralities: Flag to activate the computation of centralities.
    :type G: <networkx.classes.digraph.DiGraph>
    :type info: <dict>
    :type centralities: <boolean>
    """
    # Get transition nodes
    # See docstring...
    transition_nodes = [
        node for node, color in nx.get_node_attributes(G, 'color').items()
        if color == "blue"
    ]

    nb_nodes = len(G.nodes())
    info.update({
        'graph_nodes': nb_nodes,
        'graph_nodes_places': nb_nodes - len(transition_nodes),
        'graph_edges': len(G.edges()),
    })

    if not centralities:
        return

    # largest_connected_component = max(nx.connected_component_subgraphs(G), key=len)
    connected_components = [g for g in nx.weakly_connected_component_subgraphs(G) if len(g) > 1]
    shortest_paths = [nx.average_shortest_path_length(g) for g in connected_components]
    degree = G.degree()

    info.update({
        'centralities': {
            'strongly_connected': nx.is_strongly_connected(G),
            'weakly_connected': nx.is_weakly_connected(G),
            'max_degree': max(degree.values()),
            'min_degree': min(degree.values()),
            'average_degree': sum(degree.values()) / float(len(degree)),
            'degree': degree,
            'connected_components_number': len(shortest_paths),
            'connected_components': [len(g) for g in connected_components],
            'average_shortest_paths': sum(shortest_paths) / len(shortest_paths) if shortest_paths else "UKN",
        }
    })


def get_json_graph(G):
    """Translate Networkx graph into a dictionary ready to be dumped in a JSON file.

    .. note:: In classical JSON graph, ids of nodes are their names;
        also, their position in the array of nodes gives their numerical id,
        which is used as source or target in edges definitions.
        Here, for readability and debugging purpose, we use distinct attributes
        id and label for nodes.

    :param graph: Networkx graph.
    :type graph: <networkx.classes.digraph.DiGraph>
    :return: Serialized graph ready to be dumped in a JSON file.
    :rtype: <dict>
    """

    # Debug
    # Color attribute of the nodes
    # print(nx.get_node_attributes(G, 'color'))
    # Color attribute of the edges
    # print(nx.get_edge_attributes(G, 'color'))
    # Labels of edges
    # print(nx.get_edge_attributes(G, 'label'))
    # Conditions of edges
    ## Only edges without conditions here ?
    # print(nx.get_edge_attributes(G, 'condition'))

    json_data = dict()
    id_generator = it.count()
    nodes = list()
    edges = list()
    nodes_mapping = dict()
    # Build nodes
    for node_name, color in nx.get_node_attributes(G, 'color').items():
        node_id = next(id_generator)
        node = {
            'id': node_id,
            'label': node_name,
            'color': color,
        }
        nodes.append(node)
        nodes_mapping[node_name] = node_id

    #for edge, color in nx.get_edge_attributes(G, 'color').items():
    for ori, ext in G.edges_iter():
        # Update edge attributes
        edge = dict(G.get_edge_data(ori, ext))
        edge['source'] = nodes_mapping[ori] # get id of ori
        edge['target'] = nodes_mapping[ext] # get id of ext
        edges.append(edge)
        #{'label': G[ori][ext]['label'], 'color': G[ori][ext]['color']}
        #{'color': 'grey', 'source': 12838, 'target': 11741, 'label': '_h_1831 (0)'}

    json_data['nodes'] = nodes
    json_data['edges'] = edges

    return json_data


def export_graph(output_dir, solution, solution_index, G, *args):
    """Export a networkx graph to GraphML format.

    .. note:: Legend: See :meth:`~cadbiom_cmd.tools.graphs.build_graph`.

    :param output_dir: Output directory for GraphML files.
    :param solution: Solution string (mostly a set of frontier places).
    :param solution_index: Index of the solution in the Cadbiom result file
        (used to distinguish exported filenames).
    :param G: Networkx graph object.
    :type output_dir: <str>
    :type solution: <str>
    :type solution_index: <int> or <str>
    :type G: <networkx.classes.digraph.DiGraph>
    """

    creation_date = dt.datetime.now().strftime("%H-%M-%S")
    filename = "{}{}_{}_{}".format(
        output_dir, creation_date, solution_index, solution[:75]
    )

    # Save
    # PS: inhibitors will still have not the attribute 'color' = 'white'
    nx.write_graphml(G, filename + ".graphml")


def merge_graphs(graphs):
    """Merge graphs in the given iterable; count and add the weights to the edges
    of the final graph

    :param graphs: Networkx graph objects.
    :type graphs: <generator <networkx.classes.digraph.DiGraph>>
    :return: Networkx graph object.
    :rtype: <networkx.classes.digraph.DiGraph>
    """

    G = nx.DiGraph()

    for graph in graphs:

        missing_nodes = set(graph.nodes_iter()) - set(G.nodes_iter())
        if missing_nodes:
            # Add missing nodes in G from the current graph
            # Note: This step is mandatory even if add_edge() automatically adds
            # nodes to the graph, because some nodes can be included in any
            # transition.
            # Build a tuple (node_name, attrs)
            G.add_nodes_from((node, graph.node[node]) for node in missing_nodes)

        for ori, ext, data in graph.edges_iter(data=True):
            if G.has_edge(ori, ext):
                # Update the edge
                G[ori][ext]['weight'] += 1
            else:
                # Add the missing edge
                G.add_edge(ori, ext, attr_dict=data, weight=1)

    return G



# -*- coding: utf-8 -*-
"""Unit tests for solver results"""

#from __future__ import unicode_literals
from __future__ import print_function

# Standard imports
import pytest
import tempfile
import os

# Custom imports
from cadbiom_cmd import solution_search

@pytest.yield_fixture()
def raw_feed_output():

    # setup
    # Create the file model in /tmp/
    # Note: prevent the deletion of the file after the close() call
    fd_model = tempfile.NamedTemporaryFile(suffix='.bcx', delete=False)
    fd_model.write(
        """<model xmlns="http://cadbiom" name="bicluster">
    <CSimpleNode name="Ax" xloc="0.156748911466" yloc="0.673568818514"/>
    <CSimpleNode name="n1" xloc="0.291727140784" yloc="0.673568818514"/>
    <CSimpleNode name="C1" xloc="0.404208998548" yloc="0.618757612668"/>
    <CSimpleNode name="n2" xloc="0.405660377359" yloc="0.722289890377"/>
    <CSimpleNode name="Px" xloc="0.577083086261" yloc="0.722140333368"/>
    <CSimpleNode name="Bx" xloc="0.157474600871" yloc="0.83922046285"/>
    <CSimpleNode name="n3" xloc="0.290570367401" yloc="0.803822907245"/>
    <CSimpleNode name="n4" xloc="0.292715433748" yloc="0.906099768906"/>
    <CSimpleNode name="C2" xloc="0.409124108889" yloc="0.803710739487"/>
    <CSimpleNode name="C3" xloc="0.408336298212" yloc="0.906099768906"/>
    <transition name="" ori="Ax" ext="n1" event="h00" condition="" action="" fact_ids="[]"/>
    <transition name="" ori="n1" ext="C1" event="h0" condition="" action="" fact_ids="[]"/>
    <transition name="" ori="n1" ext="n2" event="h1" condition="C2" action="" fact_ids="[]"/>
    <transition name="" ori="n2" ext="Px" event="hlast" condition="C1 and not C3" action="" fact_ids="[]"/>
    <transition name="" ori="Bx" ext="n3" event="h2" condition="" action="" fact_ids="[]"/>
    <transition name="" ori="Bx" ext="n4" event="h4" condition="" action="" fact_ids="[]"/>
    <transition name="" ori="n3" ext="C2" event="h3" condition="" action="" fact_ids="[]"/>
    <transition name="" ori="n4" ext="C3" event="h5" condition="" action="" fact_ids="[]"/>
    </model>"""
    )
    fd_model.close()

    # fixture parameter
    yield (fd_model.name,
        (
            ['Ax Bx\n'],
            ['Ax Bx\n', '% h2\n', '% h00 h3\n', '% h0 h1\n', '% hlast\n'],
            ['4\n']
        )
    )

    # teardown
    # Delete temp files
    os.remove(fd_model.name)

@pytest.yield_fixture()
def simulate_mac_file():

    # setup
    mac_file = tempfile.NamedTemporaryFile(suffix='_mac.txt', delete=False)
    # Just frontier places
    mac_file.write("Ax Bx\nCx Dx\n")
    mac_file.close()

    # fixture parameter
    yield mac_file.name # Filename + path

    # teardown
    # Delete temp files
    os.remove(mac_file.name)

@pytest.yield_fixture()
def simulate_mac_complete_file():

    # setup
    mac_complete_file = tempfile.NamedTemporaryFile(suffix='_mac_complete.txt', delete=False)
    # 2 solutions: frontier places, then timings
    mac_complete_file.write("Ax Bx\n% h2\n% h00 h3\n% h0 h1\n% hlast\n")
    mac_complete_file.write("Cx Dx\n% h2\n% h00 h3\n")
    mac_complete_file.close()

    # fixture parameter
    yield mac_complete_file.name # Filename + path

    # teardown
    # Delete temp files
    os.remove(mac_complete_file.name)


def test_minigraph1(raw_feed_output):
    """Test the obtention of frontier places & timings on Minigraph 1

    Output must be:

        mac:
            Ax   Bx
        mac complete:
            Ax Bx
            % h2 h00  or h2
            % h3 or % h00 h3 ??
            % h0 h1
            % hlast
        mac step:
            4

    .. note:: Order of trajectories may change. (h2 h00, h3 vs h2, h3 h00)

    .. note:: This test is the equivalent of the following command line:

        cadbiom_cmd -vv debug compute_macs mini_test_publi.bcx Px --steps 10

    .. note:: The test directory is the temporary sytem folder.
    """

    fd_model_name = raw_feed_output[0]
    data = raw_feed_output[1]

    # Build params
    # See the docstring for the normal command line in shell context
    params = {
        'all_macs': False,
        'model_file': fd_model_name, # Filename + path
        'combinations': False,
        'continue': False,
        'final_prop': 'Px',
        'input_file': None,
        'inv_prop': None,
        'output': tempfile.gettempdir() + '/',
        'start_prop': None,
        'steps': 10,
        'verbose': 'debug',
        'limit': 400,
    }

    # Launch the search of the minimal accessibility condition
    solution_search.solutions_search(params)


    with open(fd_model_name[:-4] + "_Px_mac.txt", 'r') as file:
        found = [line for line in file]
        print("found mac:", found)
        assert found == data[0]

    with open(fd_model_name[:-4] + "_Px_mac_complete.txt", 'r') as file:
        found = [line for line in file]
        print("found mac_complete:", found)
        assert found == data[1]

    # Deprecated (mac_step.txt is no longer produced)
    #with open(fd_model.name[:-4] + "_Px_mac_step.txt", 'r') as file:
    #    print("found mac_step:", found)
    #    found = [line for line in file]
    #    assert found == data[2]


    # Delete temp files
    os.remove(fd_model_name[:-4] + "_Px_mac.txt")
    os.remove(fd_model_name[:-4] + "_Px_mac_complete.txt")
    #os.remove(fd_model_name[:-4] + "_Px_mac_step.txt")


def test_load_solutions(simulate_mac_file, simulate_mac_complete_file):

    from cadbiom_cmd.tools.solutions import load_solutions

    # Test solutions files parser
    # Note: this parser is not made for mac_files
    solutions = list(load_solutions(simulate_mac_file))
    expected = [('Ax Bx', []), ('Cx Dx', [])]
    print("found load_solutions mac_file", solutions)
    assert solutions == expected

    # Test solutions files parser
    solutions = list(load_solutions(simulate_mac_complete_file))
    print("found load_solutions mac_complete_file", solutions)
    expected = [
        ('Ax Bx', [['h2'], ['h00', 'h3'], ['h0', 'h1'], ['hlast']]),
        ('Cx Dx', [['h2'], ['h00', 'h3']])
    ]
    assert solutions == expected


def test_convert_solutions_to_json(raw_feed_output, simulate_mac_complete_file):

    from cadbiom_cmd.tools.solutions import load_solutions, convert_solutions_to_json
    from cadbiom_cmd.tools.models import get_transitions_from_model_file

    fd_model_name = raw_feed_output[0]

    # Get transitions from the model
    model_transitions, _ = get_transitions_from_model_file(fd_model_name)
    decomp_solutions = convert_solutions_to_json(
        load_solutions(simulate_mac_complete_file),
        model_transitions,
        conditions=True,
    )

    expected = [
        {'steps': [[{'transitions': [{'ext': 'n3', 'condition': u'', 'ori': 'Bx'}], 'event': 'h2'}],
        [{'transitions': [{'ext': 'n1', 'condition': u'', 'ori': 'Ax'}], 'event': 'h00'},
        {'transitions': [{'ext': 'C2', 'condition': u'', 'ori': 'n3'}], 'event': 'h3'}],
        [{'transitions': [{'ext': 'C1', 'condition': u'', 'ori': 'n1'}], 'event': 'h0'},
        {'transitions': [{'ext': 'n2', 'condition': u'', 'ori': 'n1'}], 'event': 'h1'}],
        [{'transitions': [{'ext': 'Px', 'condition': u'', 'ori': 'n2'}], 'event': 'hlast'}]],
        'solution': 'Ax Bx'},

        {'steps': [[{'transitions': [{'ext': 'n3', 'condition': u'', 'ori': 'Bx'}], 'event': 'h2'}],
        [{'transitions': [{'ext': 'n1', 'condition': u'', 'ori': 'Ax'}], 'event': 'h00'},
        {'transitions': [{'ext': 'C2', 'condition': u'', 'ori': 'n3'}], 'event': 'h3'}]],
        'solution': 'Cx Dx'}]

    print("found convert_solutions_to_json", decomp_solutions)
    assert decomp_solutions == expected


def test_get_all_macs(simulate_mac_file):

    from cadbiom_cmd.tools.solutions import get_mac_lines

    mac_lines = get_mac_lines(simulate_mac_file)
    print("found get_mac_lines", mac_lines)

    expected = {'Cx Dx', 'Ax Bx'}
    assert mac_lines == expected



# -*- coding: utf-8 -*-
"""Unit tests for Python API"""

from __future__ import unicode_literals
from __future__ import print_function

# Library imports
from cadbiom.models.biosignal.translators.gt_visitors import compile_cond
from cadbiom.models.guard_transitions.analyser.ana_visitors import TableVisitor
from cadbiom.models.biosignal.sig_expr import *

# Standard imports
import pytest

@pytest.fixture()
def feed_conditions():

    # {condition: (inhibitors, paths),}
    conditions = \
        {
            "((((CXCL10_CXCR3_intToMb and not(CCL11_CXCR3_intToMb)))or((CXCL10_CXCR3_intToMb and not(CXCL13_CXCR3_intToMb))))or(CXCL9_11_CXCR3_active_intToMb))or(CXCL4_CXCR3_intToMb)": ({'CCL11_CXCR3_intToMb', 'CXCL13_CXCR3_intToMb'}, [['CXCL10_CXCR3_intToMb', 'CCL11_CXCR3_intToMb'], ['CXCL10_CXCR3_intToMb', 'CXCL13_CXCR3_intToMb'], ['CXCL9_11_CXCR3_active_intToMb'], ['CXCL4_CXCR3_intToMb']]),
            "((((A and not(B)))or((A and not(C))))or(D))or(E)": ({'B', 'C'}, [['A', 'B'], ['A', 'C'], ['D'], ['E']]),
            "(A and not(B))": ({'B'}, [['A', 'B']]),
            "(((((A)or(B))or(C))or(B))or(D))or(E)": (set(), [['A'], ['B'], ['C'], ['B'], ['D'], ['E']]),
            "((((not(ATF2_JUND_macroH2A_nucl))or(Fra1_JUND_active_nucl))or(Fra1_JUN_active_nucl))or(TCF4_betacatenin_active_nucl))or(JUN_FOS_active_nucl)": ({'ATF2_JUND_macroH2A_nucl'}, [['ATF2_JUND_macroH2A_nucl'], ['Fra1_JUND_active_nucl'], ['Fra1_JUN_active_nucl'], ['TCF4_betacatenin_active_nucl'], ['JUN_FOS_active_nucl']]),
            "((A and B and C)) and (not((D and E and F and G)))": ({'E', 'D', 'G', 'F'}, [['A', 'B', 'C', 'D', 'E', 'F', 'G']]),
            "((A))or(not(cortisol))": ({'cortisol'}, [['A'], ['cortisol']]), # Test with an operator in the name of an entity
        }

    return conditions


def test_rec_tree(feed_conditions):

    from cadbiom_cmd.tools.models import rec, Reporter

    # Error Reporter
    err = Reporter()
    tvi = TableVisitor(err)
    symb_tab = tvi.tab_symb


    for condition, solution in feed_conditions.items():
        inhibitors_nodes, paths = solution

        # Get tree object from condition string
        cond_sexpr = compile_cond(condition, symb_tab, err)
        # Get all possible paths from the condition
        found_inhibitors = set()
        found_paths = rec(cond_sexpr, found_inhibitors)
        # print(found_inhibitors)
        # print(found_paths)

        assert found_paths == paths
        assert found_inhibitors == inhibitors_nodes


def test_parse_condition():

    from cadbiom_cmd.tools.models import parse_condition

    condition = "((((((STAT4__dimer___JUN_Cbp_active_nucl)or(JUN_FOSNFAT1_c_4_active_nucl))or((STAT5__dimer___active_active_nucl and Elf1)))or(IL2_IL2R_active_intToMb))or(JUN_FOSNFAT1_c_4_active_nucl))or(NFAT1_FOXP3))or(JUN_FOSNFAT1_c_4_active_nucl)"
    inhibitors = set()
    all_nodes = set(['JUN_FOSNFAT1_c_4_active_nucl', 'FOS_active_p_p_p_p_cy', 'JUN_nucl_gene', 'SRP9', 'ELK1_nucl', 'JUN_active_p_p_nucl', '14_3_3family_BAD_CaM_Ca2PLUS_CalcineurinAalpha_betaB1', 'IL2_glycosylation_exCellRegion', 'JUN_nucl', '14_3_3family', 'ERK1_2', 'STAT5_cy', 'FOS_cy', u'TGFB1__dimer___active_exCellRegion', 'NFAT1_c_4_inactive1_inactive_cy', u'SRF_nucl', 'ERK1_2_active_active_nucl', 'ERK1_2_active_active', 'ELK1_active_p_p_p_p_p_nucl', 'CaM_Ca2PLUS_CalcineurinAalpha_betaB1_active', 'FOS_cy_gene', u'PKCzeta', 'IL2Ralpha_intToMb_gene', 'IL2_IL2Ralpha_beta_gamma_JAK1_LCK_JAK3_active_intToMb', 'IL2Rgamma_JAK3_intToMb', 'STAT5__dimer___active_active_nucl', 'NFAT1_c_4_active_active_nucl', 'IL2Rbeta_JAK1_LCK_intToMb', 'SRP9_gene', 'IL2Ralpha_intToMb', 'BAD_mi', 'FOS_active_p_p_p_p_nucl'])


    found_valid_path = parse_condition(condition, all_nodes, inhibitors)

    assert found_valid_path == set([('JUN_FOSNFAT1_c_4_active_nucl',)])
    assert inhibitors == set()
